"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;
var ImplicationMethod = _interopRequireWildcard(require("../enums/ImplicationMethod"));
var AggregationMethod = _interopRequireWildcard(require("../enums/AggregationMethod"));
var DefuzzificationMethod = _interopRequireWildcard(require("../enums/DefuzzificationMethod"));
var MfCompositionType = _interopRequireWildcard(require("../enums/MfCompositionType"));
var _GenericFuzzySystem = _interopRequireDefault(require("./GenericFuzzySystem"));
var _CompositeMembershipFunction = _interopRequireDefault(require("./CompositeMembershipFunction"));
var _ConstantMembershipFunction = _interopRequireDefault(require("./ConstantMembershipFunction"));
var _MamdaniFuzzyRule = _interopRequireDefault(require("./MamdaniFuzzyRule"));
var _RuleParser = _interopRequireDefault(require("./RuleParser"));
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function _getRequireWildcardCache(nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }
function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }
/**
 * Mamdani fuzzy inference system
 */
class _default extends _GenericFuzzySystem.default {
  /**
   * Default constructor
   */
  constructor() {
    super();
    this._output = [];
    this._rules = [];
    this._implMethod = ImplicationMethod.Min;
    this._aggrMethod = AggregationMethod.Max;
    this._defuzzMethod = DefuzzificationMethod.Centroid;
  }

  /**
   * @returns {Array<FuzzyVariable>} Output linguistic variables
   */
  get output() {
    return this._output;
  }

  /**
   * @returns {Array<MamdaniFuzzyRule>} Fuzzy rules
   */
  get rules() {
    return this._rules;
  }

  /**
   * @returns {string} Implication method
   */
  get implicationMethod() {
    return this._implicationMethod;
  }

  /**
   * Implication method
   * @param {string} val
   */
  set implicationMethod(val) {
    this._implicationMethod = val;
  }

  /**
   * @returns {string} Aggregation method
   */
  get aggrMethod() {
    return this._aggrMethod;
  }

  /**
   * Aggregation method
   * @param {string} val
   */
  set aggrMethod(val) {
    this._aggrMethod = val;
  }

  /**
   * @returns {string} Defuzzification method
   */
  get defuzzMethod() {
    return this._defuzzMethod;
  }

  /**
   * Defuzzification method
   * @param {string} val
   */
  set defuzzMethod(val) {
    this._defuzzMethod = val;
  }

  /**
   * Get output linguistic variable by its name
   * @param {string} name Variable's name
   * @returns {FuzzyVariable} Found variable
   */
  outputByName(name) {
    var returnVar;
    for (var outputVar of this._output) {
      if (outputVar.name === name) {
        returnVar = outputVar;
        break;
      }
    }
    if (returnVar) {
      return returnVar;
    }
    throw new Error('Key not found');
  }

  /**
   * Create new empty rule
   * @returns {MamdaniFuzzyRule}
   */
  emptyRule() {
    return new _MamdaniFuzzyRule.default();
  }

  /**
   * Parse rule from the string
   * @param {string} rule String containing the rule
   * @returns {MamdaniFuzzyRule}
   */
  parseRule(rule) {
    return _RuleParser.default.parse(rule, this.emptyRule(), this._input, this._output);
  }

  /**
   * Calculate output values
   * @param {Map<FuzzyVariable, number>} inputValues Input values (format: variable - value)
   * @returns {Map<FuzzyVariable, number} Output values (format: variable - value)
   */
  calculate(inputValues) {
    // there should be one rule as minimum
    if (this._rules.length === 0) {
      throw new Error('There should be one rule as minimum');
    }

    // fuzzification step
    var fuzzifiedInput = this.fuzzify(inputValues);

    // evaluate the conditions
    var evaluatedConditions = this.evaluateConditions(fuzzifiedInput);

    // do implication for each rule
    var implicatedConclusions = this.implicate(evaluatedConditions);

    // aggregate the results
    var fuzzyResult = this.aggregate(implicatedConclusions);

    // defuzzify the result
    var result = this.defuzzify(fuzzyResult);
    return result;
  }

  /**
   * Evaluate conditions
   * @param {Map<FuzzyVariable, Map<FuzzyTerm, number>>} fuzzifiedInput Input in fuzzified form
   * @returns {Map<MamdaniFuzzyRule, number>} Result of evaluation
   */
  evaluateConditions(fuzzifiedInput) {
    var result = new Map();
    for (var rule of this._rules) {
      result.set(rule, this.evaluateCondition(rule.condition, fuzzifiedInput));
    }
    return result;
  }

  /**
   * Implicate rule results
   * @param {Map<MamdaniFuzzyRule, number>} conditions Rule conditions
   * @returns {Map<MamdaniFuzzyRule, IMembershipFunction>} Implicated conclusion
   */
  implicate(conditions) {
    var conclusions = new Map();
    for (var rule of conditions.keys()) {
      var compType = void 0;
      switch (this._implMethod) {
        case ImplicationMethod.Min:
          compType = MfCompositionType.Min;
          break;
        case ImplicationMethod.Production:
          compType = MfCompositionType.Prod;
          break;
        default:
          throw new Error('Internal error');
      }
      var resultMf = new _CompositeMembershipFunction.default(compType, [new _ConstantMembershipFunction.default(conditions.get(rule)), rule.conclusion.term.membershipFunction]);
      conclusions.set(rule, resultMf);
    }
    return conclusions;
  }

  /**
   * Aggregate rules
   * @param {Map<MamdaniFuzzyRile, IMembershipFunction>} conclusions Rules' results
   * @returns {Map<FuzzyVariable, IMembershipFunction} Aggregated fuzzy result
   */
  aggregate(conclusions) {
    var fuzzyResult = new Map();
    for (var outputVar of this._output) {
      var mfList = [];
      for (var rule of conclusions.keys()) {
        if (rule.conclusion.namedVar === outputVar) {
          mfList.push(conclusions.get(rule));
        }
      }
      var composType = void 0;
      switch (this._aggrMethod) {
        case AggregationMethod.Max:
          composType = MfCompositionType.Max;
          break;
        case AggregationMethod.Sum:
          composType = MfCompositionType.Sum;
          break;
        default:
          throw new Error('Internal exception');
      }
      fuzzyResult.set(outputVar, new _CompositeMembershipFunction.default(composType, mfList));
    }
    return fuzzyResult;
  }

  /**
   * Calculate crisp result for each rule
   * @param {Map<FuzzyVariable, IMembershipFunction>} fuzzyResult
   */
  defuzzify(fuzzyResult) {
    var crispResult = new Map();
    for (var fuzzyVar of fuzzyResult.keys()) {
      crispResult.set(fuzzyVar, this.defuzzifyMembershipFunction(fuzzyResult.get(fuzzyVar), fuzzyVar.min, fuzzyVar.max));
    }
    return crispResult;
  }

  /**
   *
   * @param {IMembershipFunction} mf
   * @param {number} min
   * @param {number} max
   */
  defuzzifyMembershipFunction(mf, min, max) {
    if (this._defuzzMethod === DefuzzificationMethod.Centroid) {
      var k = 50;
      var step = (max - min) / k;

      //
      // Calculate a center of gravity as integral
      //
      // let ptLeft = 0.0
      var ptCenter = 0.0;
      var ptRight = 0.0;
      var valLeft = 0.0;
      var valCenter = 0.0;
      var valRight = 0.0;
      var val2Left = 0.0;
      var val2Center = 0.0;
      var val2Right = 0.0;
      var numerator = 0.0;
      var denominator = 0.0;
      for (var i = 0; i < k; i++) {
        if (i === 0) {
          ptRight = min;
          valRight = mf.getValue(ptRight);
          val2Right = ptRight * valRight;
        }

        // ptLeft = ptRight
        ptCenter = min + step * (i + 0.5);
        ptRight = min + step * (i + 1);
        valLeft = valRight;
        valCenter = mf.getValue(ptCenter);
        valRight = mf.getValue(ptRight);
        val2Left = val2Right;
        val2Center = ptCenter * valCenter;
        val2Right = ptRight * valRight;
        numerator += step * (val2Left + 4 * val2Center + val2Right) / 3.0;
        denominator += step * (valLeft + 4 * valCenter + valRight) / 3.0;
      }
      return numerator / denominator;
    } else if (this._defuzzMethod === DefuzzificationMethod.Bisector) {
      // TODO:
      throw new Error('Bisector not supported');
    } else if (this._defuzzMethod === DefuzzificationMethod.AverageMaximum) {
      // TODO:
      throw new Error('Average Maximum not supported');
    } else {
      throw new Error('Internal exception.');
    }
  }
}
exports.default = _default;