"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;
var NameHelper = _interopRequireWildcard(require("./NameHelper"));
var HedgeType = _interopRequireWildcard(require("../enums/HedgeType"));
var OperatorType = _interopRequireWildcard(require("../enums/OperatorType"));
var _SingleCondition = _interopRequireDefault(require("./SingleCondition"));
var _FuzzyCondition = _interopRequireDefault(require("./FuzzyCondition"));
var _Conditions = _interopRequireDefault(require("./Conditions"));
var _ConditionExpression = _interopRequireDefault(require("./ConditionExpression"));
var _KeywordLexem = _interopRequireDefault(require("./KeywordLexem"));
var _TermLexem = _interopRequireDefault(require("./TermLexem"));
var _VarLexem = _interopRequireDefault(require("./VarLexem"));
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function _getRequireWildcardCache(nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }
function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }
class RuleParser {
  /**
   *
   * @param {FuzzyVariable[]} input
   * @param {NamedValue[]} output
   * @returns {Map<string, Lexem>}
   */
  static buildLexemsList(input, output) {
    var lexems = new Map();
    for (var kw of NameHelper.KEYWORDS) {
      var keywordLexem = new _KeywordLexem.default(kw);
      lexems.set(keywordLexem.text, keywordLexem);
    }
    for (var inputVar of input) {
      RuleParser.buildLexemsList2(inputVar, true, lexems);
    }
    for (var outputVar of output) {
      RuleParser.buildLexemsList2(outputVar, false, lexems);
    }
    return lexems;
  }

  /**
   *
   * @param {NamedVariable} namedVar
   * @param {boolean} input
   * @param {Map<string, Lexem>} lexems
   */
  static buildLexemsList2(namedVar, input, lexems) {
    var varLexem = new _VarLexem.default(namedVar, input);
    lexems.set(varLexem.text, varLexem);
    for (var term of namedVar.values) {
      var termLexem = new _TermLexem.default(term, input);
      if (lexems.has(termLexem.text) === false) {
        // there are no lexemsx with the same text. just insert new lexem
        lexems.set(termLexem.text, termLexem);
      } else {
        var foundLexem = lexems.get(termLexem.text);
        if (foundLexem instanceof _TermLexem.default) {
          //
          // There can be more than one terms with the same name.
          // TODO: But only if they belong to different variables.
          //
          while (foundLexem.alternative !== null) {
            foundLexem = foundLexem.alternative;
          }
          foundLexem.alternative = termLexem;
        } else {
          //
          // Only terms of different vatiables can have the same name
          //
          throw new Error("Found more than one lexems with the same name: ".concat(termLexem.text));
        }
      }
    }
  }

  /**
   *
   * @param {string} rule
   * @param {Map<string, Lexem>} lexems
   * @returns {IExpression[]}
   */
  static parseLexems(rule, lexems) {
    var expressions = [];
    var words = rule.split(' ');
    for (var word of words) {
      if (lexems.has(word)) {
        expressions.push(lexems.get(word));
      } else {
        throw new Error("Unknown identifier: ".concat(word));
      }
    }
    return expressions;
  }

  /**
   *
   * @param {IExpression[]} conditionExpression
   * @param {Map<string, Lexem>} lexems
   * @returns {IExpression[]}
   */
  static extractSingleConditions(conditionExpression, lexems) {
    var copyExpressions = conditionExpression.slice(0);
    var expressions = [];
    while (copyExpressions.length > 0) {
      if (copyExpressions[0] instanceof _VarLexem.default) {
        // parse variable
        var varLexem = copyExpressions[0];
        if (copyExpressions.length < 3) {
          throw new Error("Condition strated with '".concat(varLexem.text, "' is incorrect."));
        }
        if (varLexem.input === false) {
          throw new Error('The variable in condition part must be an input variable.');
        }

        // parse 'is' lexem

        var exprIs = copyExpressions[1];
        if (exprIs !== lexems.get('is')) {
          throw new Error("'is' keyword must go after ".concat(varLexem.text, " identifier."));
        }

        // parse 'not' lexem (if exists)
        var cur = 2;
        var not = false;
        if (copyExpressions[cur] === lexems.get('not')) {
          not = true;
          cur++;
          if (copyExpressions.length <= cur) {
            throw new Error('Error at \'not\' in condition part of the rule.');
          }
        }

        // slightly
        // somewhat
        // very
        // extremely

        //
        // Parse hedge modifier (if exists)
        //
        var hedge = HedgeType.None;
        if (copyExpressions[cur] === lexems.get('slightly')) {
          hedge = HedgeType.Slightly;
        } else if (copyExpressions[cur] === lexems.get('somewhat')) {
          hedge = HedgeType.Somewhat;
        } else if (copyExpressions[cur] === lexems.get('very')) {
          hedge = HedgeType.Very;
        } else if (copyExpressions[cur] === lexems.get('extremely')) {
          hedge = HedgeType.Extremely;
        }
        if (hedge !== HedgeType.None) {
          cur++;
          if (copyExpressions.length <= cur) {
            throw new Error("Error at '".concat(hedge, "' in condition part of the rule."));
          }
        }

        //
        // Parse term
        //
        var exprTerm = copyExpressions[cur];
        if (exprTerm instanceof _TermLexem.default === false) {
          throw new Error("Wrong identifier '".concat(exprTerm.text, "' in conditional part of the rule."));
        }
        var altLexem = exprTerm;
        var termLexem = null;
        do {
          if (!(altLexem instanceof _TermLexem.default)) {
            continue;
          }
          termLexem = altLexem;
          if (varLexem.namedVar.values.indexOf(termLexem.term) < 0) {
            termLexem = null;
            continue;
          }
        } while ((altLexem = altLexem.alternative) !== null && termLexem === null);
        if (termLexem == null) {
          throw new Error("Wrong identifier '".concat(exprTerm.text, "' in conditional part of the rule."));
        }

        //
        // Add new condition expression
        //
        var condition = new _FuzzyCondition.default(varLexem.namedVar, termLexem.term, not, hedge);
        expressions.push(new _ConditionExpression.default(copyExpressions.slice(0, cur + 1), condition));
        copyExpressions = copyExpressions.slice(cur + 1);
      } else {
        var expr = copyExpressions[0];
        if (expr === lexems.get('and') || expr === lexems.get('or') || expr === lexems.get('(') || expr === lexems.get(')')) {
          expressions.push(expr);
          copyExpressions = copyExpressions.slice(1);
        } else {
          var unknownLexem = expr;
          throw new Error("Lexem '".concat(unknownLexem.text, "' found at the wrong place in condition part of the rule."));
        }
      }
    }
    return expressions;
  }

  /**
   *
   * @param {IExpression[]} conditionExpression
   * @param {Map<string, Lexem>} lexems
   * @returns {Conditions}
   */
  static parseConditions(conditionExpression, lexems) {
    // Extract single conditions
    var expressions = RuleParser.extractSingleConditions(conditionExpression, lexems);
    if (expressions.length === 0) {
      throw new Error('No valid conditions found in conditions part of the rule.');
    }
    var cond = RuleParser.parseConditionsRecurse(expressions, lexems);

    // Return conditions
    if (cond instanceof _Conditions.default) {
      return cond;
    }
    return new _Conditions.default();
  }

  /**
   *
   * @param {IExpression[]} expressions
   * @param {Map<string, lexem>} lexems
   * @returns{number}
   */
  static findPairBracket(expressions, lexems) {
    // Assume that '(' stands at first place

    var bracketsOpened = 1;
    var closeBracket = -1;
    for (var i = 1; i < expressions.length; i++) {
      if (expressions[i] === lexems.get('(')) {
        bracketsOpened++;
      } else if (expressions[i] === lexems.get(')')) {
        bracketsOpened--;
        if (bracketsOpened === 0) {
          closeBracket = i;
          break;
        }
      }
    }
    return closeBracket;
  }

  /**
   *
   * @param {IExpression[]} expressions
   * @param {Map<string, Lexem>} lexems
   * @returns {ICondition}
   */
  static parseConditionsRecurse(expressions, lexems) {
    if (expressions.length < 1) {
      throw new Error('Empty condition found.');
    }
    if (expressions[0] === lexems.get('(') && this.findPairBracket(expressions, lexems) === expressions.length) {
      // Remove extra brackets
      return RuleParser.parseConditionsRecurse(expressions.slice(1, 1 + expressions.length - 2), lexems);
    }
    if (expressions.length === 1 && expressions[0] instanceof _ConditionExpression.default) {
      // Return single condition
      return expressions[0].condition;
    }

    // Parse list of one level conditions connected by or/and
    var copyExpressions = expressions.slice(0);
    var conds = new _Conditions.default();
    var setOrAnd = false;
    while (copyExpressions.length > 0) {
      var cond = null;
      if (copyExpressions[0] === lexems.get('(')) {
        // Find pair bracket
        var closeBracket = RuleParser.findPairBracket(copyExpressions, lexems);
        if (closeBracket === -1) {
          throw new Error('Parenthesis error');
        }
        cond = RuleParser.parseConditionsRecurse(copyExpressions.slice(1, 1 + closeBracket - 1), lexems);
        copyExpressions = copyExpressions.slice(closeBracket + 1);
      } else if (copyExpressions[0] instanceof _ConditionExpression.default) {
        cond = copyExpressions[0]._condition;
        copyExpressions = copyExpressions.slice(1);
      } else {
        throw new Error("Wrong expression in condition part at '".concat(copyExpressions[0].text));
      }

      // And condition to the list
      conds.conditionsList.push(cond);
      if (copyExpressions.length > 0) {
        if (copyExpressions[0] === lexems.get('and') || copyExpressions[0] === lexems.get('or')) {
          if (copyExpressions.length < 2) {
            throw new Error("Error at ".concat(copyExpressions[0].text, " in condition part."));
          }

          // Set and/or for conditions list
          var newOp = copyExpressions[0] === lexems.get('and') ? OperatorType.And : OperatorType.Or;
          if (setOrAnd) {
            if (conds.op !== newOp) {
              throw new Error('At the one nesting level cannot be mixed and/or operations.');
            }
          } else {
            conds.op = newOp;
            setOrAnd = true;
          }
          copyExpressions = copyExpressions.slice(1);
        } else {
          throw new Error("".concat(copyExpressions[1].text, " cannot go after ").concat(copyExpressions[0].text));
        }
      }
    }
    return conds;
  }

  /**
   *
   * @param {IExpression[]} conditionExpression
   * @param {Map<string, Lexem>} lexems
   * @returns {SingleCondition}
   */
  static parseConclusion(conditionExpression, lexems) {
    var copyExpression = conditionExpression.slice(0);

    // Remove extra brackets
    while (copyExpression.length >= 2 && copyExpression[0] === lexems.get('(') && copyExpression[copyExpression.length - 1] === lexems.get(')')) {
      copyExpression = copyExpression.slice(1, 1 + copyExpression.length - 2);
    }
    if (copyExpression.length !== 3) {
      throw new Error('Conclusion part of the rule should be in form: \'variable is term\'');
    }

    // Parse variable
    var exprVariable = copyExpression[0];
    if (!(exprVariable instanceof _VarLexem.default)) {
      throw new Error("Wrong identifier '".concat(exprVariable.text, "' in conclusion part of the rule."));
    }
    var varLexem = exprVariable;
    if (varLexem.input === true) {
      throw new Error('The variable in conclusion part must be an output variable.');
    }

    // Parse 'is' lexem
    var exprIs = copyExpression[1];
    if (exprIs !== lexems.get('is')) {
      throw new Error("'is' keyword must go after ".concat(varLexem.text, " identifier."));
    }

    // Parse term
    var exprTerm = copyExpression[2];
    if (!(exprTerm instanceof _TermLexem.default)) {
      throw new Error("Wrong identifier '".concat(exprTerm.text, "' in conclusion part of the rule."));
    }
    var altLexem = exprTerm;
    var termLexem = null;
    do {
      if (!(altLexem instanceof _TermLexem.default)) {
        // TODO: might be redundant
        continue;
      }
      termLexem = altLexem;
      if (!varLexem.namedVar.values.indexOf(termLexem.term) < 0) {
        termLexem = null;
        continue;
      }
    } while ((altLexem = altLexem.alternative) !== null && termLexem === null);
    if (termLexem === null) {
      throw new Error("Wrong identifier '".concat(exprTerm.text, "' in conclusion part of the rule."));
    }

    // Return fuzzy rule's conclusion
    return new _SingleCondition.default(varLexem.namedVar, termLexem.term, false);
  }

  /**
   *
   * @param {string} rule
   * @param {GenericFuzzyRule} emptyRule
   * @param {FuzzyVariable[]} input
   * @param {NamedVariable[]} output
   * @returns {GenericFuzzyRule}
   */
  static parse(rule, emptyRule, input, output) {
    if (rule.length === 0) {
      throw new Error('Rule cannot be empty.');
    }

    // Surround brakes with spaces, remove double spaces
    var sb = [];
    for (var ch of rule) {
      if (ch === ')' || ch === '(') {
        if (sb.length > 0 && sb[sb.length - 1] === ' ') {
          // Do not duplicate spaces
        } else {
          sb.push(' ');
        }
        sb.push(ch);
        sb.push(' ');
      } else {
        if (ch === ' ' && sb.length > 0 && sb[sb.length - 1] === ' ') {
          // do not duplicate spaces
        } else {
          sb.push(ch);
        }
      }
    }

    // Remove spaces
    var prepRule = sb.join('').trim();

    // Build lexems dictionary
    var lexemsDict = RuleParser.buildLexemsList(input, output);

    // At first we parse lexems
    var expressions = RuleParser.parseLexems(prepRule, lexemsDict);
    if (expressions.length === 0) {
      throw new Error('No valid identifiers found.');
    }

    // Find condition & conclusion parts part
    if (expressions[0] !== lexemsDict.get('if')) {
      throw new Error('\'if\' should be the first identifier.');
    }
    var thenIndex = -1;
    for (var i = 1; i < expressions.length; i++) {
      if (expressions[i] === lexemsDict.get('then')) {
        thenIndex = i;
        break;
      }
    }
    if (thenIndex === -1) {
      throw new Error('\'then\' identifier not found.');
    }
    var conditionLen = thenIndex - 1;
    if (conditionLen < 1) {
      throw new Error('Condition part of the rule not found.');
    }
    var conclusionLen = expressions.length - thenIndex - 1;
    if (conclusionLen < 1) {
      throw new Error('Conclusion part of the rule not found.');
    }
    var conditionExpressions = expressions.slice(1, 1 + conditionLen);
    var conclusionExpressions = expressions.slice(thenIndex + 1, thenIndex + 1 + conclusionLen);
    var conditions = RuleParser.parseConditions(conditionExpressions, lexemsDict);
    var conclusion = RuleParser.parseConclusion(conclusionExpressions, lexemsDict);
    emptyRule.condition = conditions;
    emptyRule.conclusion = conclusion;
    return emptyRule;
  }
}
exports.default = RuleParser;