"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;
var _GenericFuzzySystem = _interopRequireDefault(require("./GenericFuzzySystem"));
var _SugenoFuzzyRule = _interopRequireDefault(require("./SugenoFuzzyRule"));
var _LinearSugenoFunction = _interopRequireDefault(require("./LinearSugenoFunction"));
var _RuleParser = _interopRequireDefault(require("./RuleParser"));
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
/**
 * Sugeno fuzzy inference system
 */
class _default extends _GenericFuzzySystem.default {
  /**
   * Default constructor
   */
  constructor() {
    super();
    this._output = [];
    this._rules = [];
  }

  /**
   * @returns {Array<SugenoVariable>} Output linguistic variables
   */
  get output() {
    return this._output;
  }

  /**
   * @returns {Array<SugenoFuzzyRule>} Fuzzy rules
   */
  get rules() {
    return this._rules;
  }

  /**
   * Get output linguistic variable by its name
   * @param {string} name Variable's name
   * @returns {SugenoVariable} Found variable
   */
  outputByName(name) {
    var returnVar;
    for (var outputVar of this._output) {
      if (outputVar.name === name) {
        returnVar = outputVar;
        break;
      }
    }
    if (returnVar) {
      return returnVar;
    }
    throw new Error('Key not found');
  }

  /**
   * Create new empty rule
   * @returns {SugenoFuzzyRule}
   */
  emptyRule() {
    return new _SugenoFuzzyRule.default();
  }

  /**
   * Use this method to create a linear function for the Sugeno fuzzy system
   * @param {string} name Name of the function
   * @param {Array|Map<FuzzyVariable, number>} coeffs List of coefficients. List length must be less or equal to the input length
   * @param {number} [constValue]
   * @returns {LinearSugenoFunction} Created function
   */
  createSugenoFunction(name, coeffs, constValue) {
    return new _LinearSugenoFunction.default(name, this._input, coeffs, constValue);
  }

  /**
   * Use this method to create rule by its textual representation
   * @param {string} rule Rule in text form
   * @returns {SugenoFuzzyRule} Created rule
   */
  parseRule(rule) {
    return _RuleParser.default.parse(rule, this.emptyRule(), this._input, this._output);
  }

  /**
   * Calculate output of fuzzy system
   * @param {Map<FuzzyVariable, number>} inputValues Input values
   * @returns {Map<SugenoVariable, number} Output values
   */
  calculate(inputValues) {
    // there should be one rule as minimum
    if (this._rules.length === 0) {
      throw new Error('There should be one rule as minimum');
    }

    // fuzzification step
    var fuzzifiedInput = this.fuzzify(inputValues);

    // evaluate the conditions
    var ruleWeights = this.evaluateConditions(fuzzifiedInput);

    // functions evaluation
    var functionsResult = this.evaluateFunctions(inputValues);

    // combine output
    var result = this.combineResult(ruleWeights, functionsResult);
    return result;
  }

  /**
   * Evaluate conditions
   * @param {Map<FuzzyVariable, Map<FuzzyTerm, number>>} fuzzifiedInput Input in fuzzified form
   * @returns {Map<SugenoFuzzyRule, number>} Result of evaluation
   */
  evaluateConditions(fuzzifiedInput) {
    var result = new Map();
    for (var rule of this._rules) {
      result.set(rule, this.evaluateCondition(rule.condition, fuzzifiedInput));
    }
    return result;
  }

  /**
   * Calculate functions' results
   * @param {Map<FuzzyVariable, number>} inputValues Input values
   * @returns {Map<SugenoVariable, Map<LinearSugenoFunction, number>>} Results
   */
  evaluateFunctions(inputValues) {
    var result = new Map();
    for (var outputVar of this._output) {
      var varResult = new Map();
      for (var func of outputVar.functions) {
        varResult.set(func, func.evaluate(inputValues));
      }
      result.set(outputVar, varResult);
    }
    return result;
  }

  /**
   * Combine results of functions and rule evaluation
   * @param {Map<SugenoFuzzyRule, number>} ruleWeights Rule weights (results of evaluation)
   * @param {Map<SugenoVariable, Map<LinearSugenoFunction, number>>} functionResults Result of functions evaluation
   * @returns Result of calculations
   */
  combineResult(ruleWeights, functionResults) {
    var numerators = new Map();
    var denominators = new Map();
    var results = new Map();

    // Calculate numerator and denominator separately for each output
    for (var outputVar of this._output) {
      numerators.set(outputVar, 0.0);
      denominators.set(outputVar, 0.0);
    }
    for (var rule of ruleWeights.keys()) {
      var conclusionVar = rule.conclusion.namedVar;
      var z = functionResults.get(conclusionVar).get(rule.conclusion.term);
      var w = ruleWeights.get(rule);
      numerators.set(conclusionVar, numerators.get(conclusionVar) + z * w);
      denominators.set(conclusionVar, denominators.get(conclusionVar) + w);
    }

    // calculate the fractions
    for (var _outputVar of this._output) {
      if (denominators.get(_outputVar) === 0.0) {
        results.set(_outputVar, 0.0);
      } else {
        results.set(_outputVar, numerators.get(_outputVar) / denominators.get(_outputVar));
      }
    }
    return results;
  }
}
exports.default = _default;